from pyqchem.structure import Structure
from pyqchem.parsers.common import read_input_structure
import numpy as np
import re


def parse_molecule(opt_section, charge=0, multiplicity=0):

    num = opt_section.find('Coordinates')
    if num < 0:
        num = opt_section.find('Standard Nuclear Orientation')
        coordinates_section = opt_section[num:].split('\n')[3:]
    else:
        coordinates_section = opt_section[num:].split('\n')[2:]

    # check number of atoms
    n_atoms = 0
    for i in coordinates_section:
        if len(i.split()) != 5:
            break
        n_atoms += 1

    coordinates_section = coordinates_section[:n_atoms]

    symbols = [line.split()[1] for line in coordinates_section]
    coordinates = [line.split()[2:] for line in coordinates_section]

    molecule = Structure(coordinates=np.array(coordinates, dtype=float).tolist(),
                         symbols=symbols,
                         charge=charge,
                         multiplicity=multiplicity)

    return molecule


def basic_optimization(output, print_data=False):

    data_dict = {}

    # Molecule
    structure = read_input_structure(output)

    step_s2 = None
    # Optimization steps
    optimization_steps = []
    list_iterations = [l.end() for l in re.finditer('Optimization Cycle', output, re.IGNORECASE)]
    for ini, fin in zip(list_iterations, list_iterations[1:] + [len(output)]):
        step_section = output[ini:fin]
        enum = step_section.find('Coordinates (Angstroms)')

        step_molecule = parse_molecule(step_section, structure.charge, structure.multiplicity)

        enum = step_section.find('Energy is')
        step_energy = float(step_section[enum: enum+50].split()[2])
        enum = step_section.find('      Gradient')
        step_gradient = float(step_section[enum: enum+50].split()[1])
        enum = step_section.find('      Displacement')
        step_displacement = float(step_section[enum: enum+50].split()[1])

        enum = step_section.find('<S^2>')
        if enum > 0:
            step_s2 = float(step_section[enum: enum+50].split()[2])

        optimization_steps.append({'molecule': step_molecule,
                                   'energy': step_energy,
                                   'gradient': step_gradient,
                                   'displacement': step_displacement})

    data_dict['optimization_steps'] = optimization_steps

    # Optimization Convergence
    enum = output.find('**  OPTIMIZATION CONVERGED  **')
    if enum > 0:
        ne = output[enum-200:enum].find('Final energy')

        final_energy = float(output[ne+enum-200: enum].split()[3])
        optimization_section = output[enum:]

        optimized_molecule = parse_molecule(optimization_section, structure.charge, structure.multiplicity)

        data_dict['optimized_molecule'] = optimized_molecule
        data_dict['energy'] = final_energy
        data_dict['s2'] = step_s2

    return data_dict





from pyqchem.qc_input import QchemInput
from pyqchem.errors import ParserError, OutputError
from pyqchem.utils import get_sdm
from subprocess import Popen, PIPE
from pathlib import Path
import os, shutil, sys
import numpy as np
import hashlib
import pickle
import warnings
import copy


if sys.version_info[0] < 3 or sys.platform in ["win32", "cygwin"] or os.getenv('PYQCHEM_CACHE') == '1':
    # For python 2.x or Windows use pickle based cache system
    from pyqchem.cache import SimpleCache as CacheSystem
    warnings.warn('Using SimpleCache')
else:
    # For python 3.x use SQL Database based cache system
    from pyqchem.cache import SqlCache as CacheSystem


# Backwards Compatibility
def redefine_calculation_data_filename(filename, compress=False):
    cache = CacheSystem(filename=filename, compress=compress)


# Check if calculation finished ok
def finish_ok(output):
    return output[-1000:].find('Thank you very much for using Q-Chem') != -1


# Return qchem version
def get_version_output(output):

    class QChemVersion:
        def __init__(self, string):

            string_version = string.split()[1]
            self._major = string_version.split('.')[0]
            self._minor = string_version.split('.')[1]

            string_branch = string.split()[2]
            self._devel = True if '(devel)' in string_branch else False

        def __str__(self):
            dev = 'dev' if self.is_development else ''
            return '{}.{} {}'.format(self.major, self.minor, dev)

        def __eq__(self, other):

            """
            Here put the logic for more sophisticated comparison
            between versions

            :param other: string/QchemVersion
            :return:
            """

            if isinstance(other, QChemVersion):
                return True if self.__str__() == other.__str__() else False

            o_major = other.split('.')[0]
            o_minor = other.split('.')[1]

            if int(o_major) == self.major:

                # handle expresions like 2.3+
                if '+' in o_minor[-1]:
                    if self.minor >= int(o_minor[:-1]):
                        return True
                    else:
                        return False

                if int(o_minor) == self.minor:
                    return True

            return False

        @property
        def major(self):
            return int(self._major)

        @property
        def minor(self):
            return int(self._minor)

        @property
        def is_development(self):
            return self._devel

    index = output[:500].find('\n Q-Chem')
    string = output[index: index + 30]

    return QChemVersion(string)


def get_compatibility_list_from_parser(parser):
    docstring = parser.__doc__
    if docstring is None:
        return None

    lines = docstring.split('\n')
    for line in lines:
        if 'compatibility' in line.lower():
            try:
                return [version.strip() for version in line.split(':')[1].split(',')]
            except IndexError:
                continue

    return None


# Layer of compatibility with old version
def create_qchem_input(*args, **kwargs):
    return QchemInput(*args, **kwargs)


def parse_output(get_output_function):
    """
    to be deprecated

    :param get_output_function:
    :return: parsed output
    """

    cache = CacheSystem()

    def func_wrapper(*args, **kwargs):
        parser = kwargs.pop('parser', None)
        parser_parameters = kwargs.pop('parser_parameters', {})
        store_output = kwargs.pop('store_output', None)

        force_recalculation = kwargs.pop('force_recalculation', False)

        if parser is not None:
            hash_p = (args[0], parser.__name__)
            if hash_p in cache.calculation_data and not force_recalculation:
                print('already calculated. Skip')
                return cache.calculation_data[hash_p]

        output, err = get_output_function(*args, **kwargs)

        if store_output is not None:
            with open('{}'.format(store_output), 'w') as f:
                f.write(output)

        if len(err) > 0:
            print(output[-800:])
            print(err)
            raise Exception('q-chem calculation finished with error')

        if parser is None:
            return output

        parsed_output = parser(output, **parser_parameters)

        cache.calculation_data[hash_p] = parsed_output
        with open(cache._calculation_data_filename, 'wb') as output:
            pickle.dump(cache.calculation_data, output, protocol=pickle.DEFAULT_PROTOCOL)

        return parsed_output

    return func_wrapper


def local_run(input_file_name, work_dir, fchk_file, use_mpi=False, processors=1):
    """
    Run Q-Chem locally

    :param input_file_name: Q-Chem input file in plain text format
    :param work_dir:  Scratch directory where calculation run
    :param fchk_file: filename of fchk
    :param use_mpi: use mpi instead of openmp

    :return: output, err: Q-Chem standard output and standard error
    """

    if not use_mpi:
        os.environ["QCTHREADS"] = "{}".format(processors)
        os.environ["OMP_NUM_THREADS"] = "{}".format(processors)
        os.environ["MKL_NUM_THREADS"] = "1"

    os.environ["GUIFILE"] = fchk_file
    qc_dir = os.getenv('QC')
    exe_dir = os.getenv('QC_EXE_DIR') if 'QC_EXE_DIR' in os.environ else 'exe'

    binary = Path(qc_dir).joinpath(exe_dir).joinpath('qcprog.exe')
    command = [binary, Path(work_dir).joinpath(input_file_name), Path(work_dir)]

    qchem_process = Popen(command, stdout=PIPE, stdin=PIPE, stderr=PIPE, cwd=work_dir)
    (output, err) = qchem_process.communicate()
    qchem_process.wait()
    output = output.decode(errors='ignore')
    err = err.decode()

    return output, err

def local_run_stream(input_file_name, work_dir, fchk_file, use_mpi=False, processors=1, print_stream=True):
    """
    Run Q-Chem locally

    :param input_file_name: Q-Chem input file in plain text format
    :param work_dir:  Scratch directory where calculation run
    :param fchk_file: filename of fchk
    :param use_mpi: use mpi instead of openmp
    :param print_stream: set True to print output stream during execution

    :return: output, err: Q-Chem standard output and standard error
    """

    if not use_mpi:
        os.environ["QCTHREADS"] = "{}".format(processors)
        os.environ["OMP_NUM_THREADS"] = "{}".format(processors)
        os.environ["MKL_NUM_THREADS"] = "1"

    os.environ["GUIFILE"] = fchk_file
    qc_dir = os.getenv('QC')

    exe_dir = os.getenv('QC_EXE_DIR') if 'QC_EXE_DIR' in os.environ else 'exe'
    binary = Path(qc_dir).joinpath(exe_dir).joinpath('qcprog.exe')
    command = [binary, Path(work_dir).joinpath(input_file_name), Path(work_dir)]

    qchem_process = Popen(command, stdout=PIPE, stdin=PIPE, stderr=PIPE, cwd=work_dir)

    output = ''
    err = ''
    while True:
        line_out = qchem_process.stdout.readline()
        line_err = qchem_process.stderr.readline()

        if not line_out and not line_err:
            break

        if print_stream:
            print(line_out.strip().decode(errors='ignore'))

        sys.stdout.flush()

        output += line_out.decode(errors='ignore')
        err += line_err.decode(errors='ignore')

    return output, err


def remote_run(input_file_name, work_dir, fchk_file, remote_params, use_mpi=False, processors=1):
    """
    Run Q-Chem remotely

    :param input_file: Q-Chem input file in plain text format
    :param work_dir:  Scratch directory where calculation run
    :param fchk_file: filename of fchk
    :param remote_params: connection parameters for paramiko
    :param use_mpi: use mpi instead of openmp

    :return: output, err: Q-Chem standard output and standard error
    """
    import paramiko
    remote_params = copy.deepcopy(remote_params)

    # get precommands
    commands = remote_params.pop('precommand', [])
    remote_scratch = remote_params.pop('remote_scratch', None)

    # Setup SSH connection
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(**remote_params)

    ssh.get_transport()
    sftp = ssh.open_sftp()
    print('connected to {}..'.format(remote_params['hostname']))

    # Define temp remote dir
    _, stdout, _ = ssh.exec_command('pwd', get_pty=True)

    if remote_scratch is None:
        remote_scratch = stdout.read().decode().strip('\n').strip('\r')

    remote_dir = os.path.join(remote_scratch, 'temp_pyqchem_remote')

    # Create temp directory in remote machine
    try:
        sftp.mkdir(remote_dir)
    except OSError:
        pass
    sftp.chdir(remote_dir)

    # Copy all files in local workdir to remote machine
    file_list = os.listdir(work_dir)
    for file in file_list:
        sftp.put(os.path.join(work_dir, file), '{}'.format(file))

    flag = '-np' if use_mpi else '-nt'

    # Define commands to run Q-Chem in remote machine
    commands += ['cd  {}'.format(remote_dir),  # go to remote work dir
                 'qchem {} {} {}'.format(flag, processors, input_file_name)]  # run qchem

    # Execute command in remote machine
    stdin, stdout, stderr = ssh.exec_command('bash -l -c "{}"'.format(';'.join(commands)), get_pty=True)

    # Reformat output/error files
    output = ''.join(stdout.readlines())
    error = ''.join(stderr.readlines())

    # get files and remove them from remote server
    for file in sftp.listdir():
        sftp.get(os.path.join(remote_dir, file), os.path.join(work_dir, file))
        sftp.remove(os.path.join(remote_dir, file))
    sftp.rmdir(remote_dir)

    sftp.close()
    ssh.close()

    # Rename fchk file to match expected name
    if input_file_name + '.fchk' in os.listdir(work_dir):
        os.rename(os.path.join(work_dir, input_file_name + '.fchk'), os.path.join(work_dir, fchk_file))

    return output, error


def generate_additional_files(input_qchem, work_dir):
    """
    Generate additional files on scratch (work dir) for special calculations

    :param input_qchem: QChem input object
    :param work_dir: scratch directory
    """

    # handle custom guess
    if input_qchem.mo_coefficients is not None:
        input_qchem.store_mo_file(work_dir)

    if input_qchem.scf_density is not None:
        input_qchem.store_density_file(work_dir)

    # set scf energy if skip_scfman (to not break)
    # TODO: now SCF energy is set to zero. This may not work for all features.
    if input_qchem._skip_scfman:
        if input_qchem.mo_coefficients is None:
            raise Exception('Explicit MO guess has to be provided for scf_skip')
        input_qchem.store_energy_file(work_dir)

    # Write hessian
    if input_qchem.hessian is not None:
        input_qchem.store_hessian_file(work_dir)

    # write RAS-GUESS
    if input_qchem.ras_guess is not None:
        input_qchem.store_ras_guess_file(work_dir)


def retrieve_additional_files(input_qchem, data_fchk, work_dir, scratch_read_level=0):
    """
    retrieve data from files in scratch data (on development, currently for test only)

    :param input_qchem: QChem input object
    :param data_fchk: FCHK parsed dictionary
    :param work_dir: scratch directory
    :param scratch_read_level: defines what data to retrieve
    :return: dictionary with additional data
    """

    additional_data = {}

    natom = len(input_qchem.molecule.get_coordinates())
    file_list = os.listdir(work_dir)

    # OLD_DIMENSIONS
    if '819.0' in file_list:
        with open(work_dir + '819.0', 'r') as f:
            data = np.fromfile(f, dtype=np.int32)
            norb_alpha, norb_beta = data[0:2]
            norb = norb_alpha
            nbas = norb  # assumption
    else:
        norb = np.shape(data_fchk['coefficients']['alpha'])[0]
        nbas = np.shape(data_fchk['coefficients']['alpha'])[1]


    # MO_COEFS (Already in fchk) in internal order
    if '53.0' in file_list and 'coefficients' in data_fchk:
        with open(work_dir + '53.0', 'r') as f:
            data = np.fromfile(f, dtype=float)
            mo_alpha = data[:norb*nbas].reshape(-1, norb).tolist()
            mo_beta = data[norb*nbas: 2*norb_beta*nbas].reshape(-1, norb_beta).tolist()
            # additional_data['coefficients_internal'] = {'alpha': mo_alpha, 'beta': mo_beta}

            # obtain the order indices between fchk order and Q-Chem internal order of basis functions
            diff_square = get_sdm(data_fchk['coefficients']['alpha'], mo_alpha)

            # get non-repeating indices
            indices = []
            for row in diff_square.T:
                for i in np.argsort(row):
                    if i not in indices:
                        indices.append(int(i))
                        break

            # indices = np.argmin(diff_square, axis=0).tolist()

            # store q-chem index order for later use (e.g  guess)
            data_fchk['coefficients']['qchem_order'] = indices
    else:
        indices = list(range(nbas))

    # FOCK_MATRIX
    if '58.0' in file_list:
        with open(work_dir + '58.0', 'r') as f:
            data = np.fromfile(f, dtype=float)
            fock_alpha = data[:nbas*nbas].reshape(-1, nbas)
            fock_beta = data[nbas*nbas: 2*nbas*nbas].reshape(-1, nbas)

            # set basis functions in fchk order
            fock_alpha = fock_alpha[:, indices]
            fock_alpha = fock_alpha[indices, :]
            fock_beta = fock_beta[:, indices]
            fock_beta = fock_beta[indices, :]

            additional_data['fock_matrix'] = {'alpha': fock_alpha.tolist(), 'beta': fock_beta.tolist()}

    if scratch_read_level == -1:
        # FILE_ENERGY (Not really worth to read it)
        if '99.0' in file_list:
            with open(work_dir + '99.0', 'r') as f:
                data = np.fromfile(f, dtype=float)

        # FILE_DENSITY_MATRIX (Already in fchk)
        if '54.0' in file_list:
            with open(work_dir + '54.0', 'r') as f:
                data = np.fromfile(f, dtype=float)
                density_alpha = data[:nbas*nbas].reshape(-1, nbas)
                density_beta = data[nbas*nbas: 2*nbas*nbas].reshape(-1, nbas)
                # set basis functions in fchk order
                density_alpha = density_alpha[:, indices]
                density_alpha = density_alpha[indices, :]
                density_beta = density_beta[:, indices]
                density_beta = density_beta[indices, :]
                additional_data['scf_density_internal'] = {'alpha': density_alpha.tolist(), 'beta': density_beta.tolist()}

    # HESSIAN_MATRIX
    if '132.0' in file_list:
        with open(work_dir + '132.0', 'r') as f:
            data = np.fromfile(f, dtype=float)
            hessian = data.reshape(-1, natom*3)
            additional_data['hessian'] = hessian.tolist()

    # AO_INTS_DEBUG
    if '21.0' in file_list:
        with open(work_dir + '21.0', 'r') as f:
            data = np.fromfile(f, dtype=float)
            ao_integrals = data.reshape(-1, nbas, nbas, nbas)

            # set basis functions in fchk order
            ao_integrals = ao_integrals[:, :, :, indices]
            ao_integrals = ao_integrals[:, :, indices, :]
            ao_integrals = ao_integrals[:, indices, :, :]
            ao_integrals = ao_integrals[indices, :, :, :]

            additional_data['ao_integrals'] = ao_integrals.tolist()

    if scratch_read_level > 0:
        # FILE_RAS_AMP
        if '704.0' in file_list:
            with open(work_dir + '705.0', 'r') as f:
                ras_energies = np.fromfile(f, dtype=float)
                n_ras_roots = len(ras_energies)

            with open(work_dir + '704.0', 'r') as f:
                data = np.fromfile(f, dtype=float)
                ras_amplitudes = data.reshape(n_ras_roots, -1)
                additional_data['ras_amplitudes'] = ras_amplitudes.tolist()

    return additional_data


def get_output_from_qchem(input_qchem,
                          processors=1,
                          use_mpi=False,
                          scratch=None,
                          read_fchk=False,  # to be deprecated
                          return_electronic_structure=False,
                          parser=None,
                          parser_parameters=None,
                          force_recalculation=False,
                          fchk_only=False,
                          store_full_output=False,
                          delete_scratch=True,
                          remote=None,
                          scratch_read_level=0):

    """
    Runs qchem and returns the output in the following format:

    1) If return_electronic_structure is requested:
        [output, parsed_fchk]
    2) If return_electronic_structure is not requested:
        [output]

    Note: if parser is set then output contains a dictionary with the parsed info
          else output contains the q-chem output in plain text

    :param input_qchem: QcInput object containing the Q-Chem input
    :param processors: number of threads/processors to use in the calculation
    :param use_mpi: If False use OpenMP (threads) else use MPI (processors)
    :param scratch: Full Q-Chem scratch directory path. If None read from $QCSCRATCH
    :param return_electronic_structure: if True, returns the parsed FCHK file containing the electronic structure
    :param read_fchk: same as return_electronic_structure (to be deprecated)
    :param parser: function to use to parse the Q-Chem output
    :param parser_parameters: additional parameters that parser function may have
    :param force_recalculation: Force to recalculate even identical calculation has already performed
    :param fchk_only: If true, returns electronic structure data from cache ignoring output (to be deprecated)
    :param remote: dictionary containing the data for remote calculation (beta)
    :param store_full_output: store full output in plain text in pkl file
    :param delete_scratch: delete all scratch files when calculation is finished

    :return: output [, electronic_structure]
    """
    from pyqchem.parsers.parser_fchk import parser_fchk
    cache = CacheSystem()

    # back-compatibility layer
    if read_fchk:
        warnings.warn("'read_fchk' will be deprecated, use return_electronic_structure instead",
                      DeprecationWarning, stacklevel=2)
        return_electronic_structure = read_fchk

    if fchk_only:
        warnings.warn("'fchk_only' option will be deprecated",
                      DeprecationWarning, stacklevel=2)

    # Always generate fchk
    if input_qchem.gui is None or input_qchem.gui < 1:
        input_qchem.gui = 2

    if scratch is None:
        scratch = os.getenv('QCSCRATCH')
        if scratch is None:
            warnings.warn('QCSCRATCH environment variable not defined, using workdir')
            scratch = '.'

    work_dir = '{}/pyqchem_{}/'.format(scratch, os.getpid())

    try:
        os.mkdir(work_dir)
    except OSError:
        pass

    # check if parameters is None
    if parser_parameters is None:
        parser_parameters = {}

    # check if full output is stored
    output = cache.retrieve_calculation_data(input_qchem, 'fullout')

    #output, err = cache.calculation_data[(hash(input_qchem), 'fullout')] if (hash(input_qchem), 'fullout') in cache.calculation_data else [None, None]
    elect_struct_data = cache.retrieve_calculation_data(input_qchem, 'fchk')

    # check if repeated calculation
    if not force_recalculation and not store_full_output:  # store_full_output always force re-parsing
        if parser is not None:
            parsed_data = cache.retrieve_calculation_data(hash(input_qchem), parser.__name__)
            if parsed_data is not None:
                if return_electronic_structure:
                    return parsed_data, elect_struct_data
                else:
                    return parsed_data
        else:
            if fchk_only and elect_struct_data is not None:
                return output, elect_struct_data

    # temp filenames generated in temp directory
    fchk_filename = 'qchem_temp_{}.fchk'.format(os.getpid())
    temp_filename = 'qchem_temp_{}.inp'.format(os.getpid())

    # generate the input in TXT form
    input_txt = input_qchem.get_txt()
    qchem_input_file = open(os.path.join(work_dir, temp_filename), mode='w')
    qchem_input_file.write(input_txt)
    qchem_input_file.close()

    # generate extra files in calculation directory
    generate_additional_files(input_qchem, work_dir)

    # Q-Chem calculation
    if output is None or force_recalculation is True:
        if remote is None:
            # print('local:')
            output, err = local_run(temp_filename, work_dir, fchk_filename, use_mpi=use_mpi, processors=processors)
        else:
            # print('Remote:')
            output, err = remote_run(temp_filename, work_dir, fchk_filename, remote, use_mpi=use_mpi, processors=processors)

        if not finish_ok(output):
            raise OutputError(output, err)

        # parse fchk file & and additional scratch dir files
        if not os.path.isfile(os.path.join(work_dir, fchk_filename)):
            warnings.warn('fchk not found! something may be wrong in calculation')
        else:
            with open(os.path.join(work_dir, fchk_filename)) as f:
                fchk_txt = f.read()

            elect_struct_data = parser_fchk(fchk_txt)
            elect_struct_data.update(retrieve_additional_files(input_qchem,
                                                               elect_struct_data,
                                                               work_dir,
                                                               scratch_read_level))
            cache.store_calculation_data(input_qchem, 'fchk', elect_struct_data)

        if store_full_output:
            cache.store_calculation_data(input_qchem, 'fullout', output)

    if parser is not None:

        # Check parser compatibility
        version = get_version_output(output)
        compatibility_list = get_compatibility_list_from_parser(parser)
        if compatibility_list is not None:
            if version not in compatibility_list:
                warnings.warn('Parser "{}" may not be compatible with Q-Chem {}'.format(parser.__name__, version))

        # minimum functionality for parser error capture
        try:
            output = parser(output, **parser_parameters)
        except:
            raise ParserError(parser.__name__, 'Undefined error', output)

        cache.store_calculation_data(input_qchem, parser.__name__, output)

    if delete_scratch:
        shutil.rmtree(work_dir)

    if return_electronic_structure:
        return output, elect_struct_data
    else:
        return output


def get_input_hash(data):
    return hashlib.md5(data.encode()).hexdigest()




__author__ = 'Abel Carreras'

import re
import operator
import warnings

import numpy as np
from pyqchem.parsers.common import read_basic_info, get_rasci_occupations_list
from pyqchem.parsers.common import search_bars, standardize_vector, read_input_structure
from pyqchem.parsers.common import float_asterisk as float


def _read_simple_matrix(header, output, maxchar=10000, foot='-------'):
    matrix_list = []
    for m in re.finditer(header, output):
        section_state = output[m.end():m.end() + maxchar]  # 10000: assumed to max of section
        section_state = section_state[:section_state.find(foot)]
        dim = len(section_state.split('\n')[1].split())
        matrix = section_state.split('\n')[1:dim + 1]
        matrix = [[float(n) for n in l.split()] for l in matrix]
        matrix_list.append(matrix)

    return matrix_list


def _read_soc_matrix(lines, dimensions):

    col_per_line = 5
    matrix = []
    for ib in range(dimensions[0]):
        real = []
        complex = []
        for j in range((dimensions[1] - 1) // col_per_line + 1):
            real += lines[j*dimensions[0] + 1 * (j+1) + ib][11:].split()[0::2]
            complex += lines[j*dimensions[0] + 1 * (j+1) +ib][11:].split()[1::2]

        row = [float(r) + float(c[:-1]) * 1j for r, c in zip(real, complex)]
        matrix.append(row)

    return matrix


def _complete_interstate_pairs(interstate_dict):
    additional_items = {}
    for key, value in interstate_dict.items():
        if not key[::-1] in interstate_dict:
            dict_entry = {}
            for k, v in interstate_dict[key].items():
                if k == 'state_a':
                    dict_entry.update({k: key[1]})
                elif k == 'state_b':
                    dict_entry.update({k: key[0]})
                elif k == 'angular_momentum':
                    dict_entry.update({k: np.conjugate(v).tolist()})
                elif k == '1e_soc_mat':
                    dict_entry.update({k: np.conjugate(v).T.tolist()})
                elif k == 'hso_l-':
                    dict_entry.update({k: (-np.array(v)).tolist()})
                elif k == 'hso_l+':
                    dict_entry.update({k: (-np.array(v)).tolist()})
                elif k == '2e_soc_mat':
                    dict_entry.update({k: np.conjugate(v).T.tolist()})
                elif k == 'total_soc_mat':
                    dict_entry.update({k: np.conjugate(v).T.tolist()})
                else:
                    dict_entry.update({k: v})
            additional_items.update({key[::-1]: dict_entry})

    interstate_dict.update(additional_items)


def parser_rasci(output):
    """
    Parser for RAS-CI calculations
    Include:
    - Diabatization scheme data
    - Structure
    - Adiabatic states
    - SOC

    :param output:
    :return:
    """

    data_dict = {}

    # Molecule
    data_dict['structure'] = read_input_structure(output)
    n_atoms = data_dict['structure'].get_number_of_atoms()

    # basic info
    enum = output.find('Nuclear Repulsion Energy')
    basic_data = read_basic_info(output[enum:enum + 5000])

    # scf_energy
    enum = output.find('SCF   energy in the final basis set')
    scf_energy = float(output[enum:enum+100].split()[8])

    data_dict['scf_energy'] = scf_energy
    # total energy
    # enum = output.find('Total energy in the final basis set')
    # total_energy = float(output[enum:enum+100].split()[8])

    # RASCI dimensions
    ini_section = output.find('RAS-CI Dimensions')
    end_section = search_bars(output, from_position=ini_section, bar_type=r'\*\*\*')[0]
    dimension_section = output[ini_section: end_section]

    enum = dimension_section.find('Doubly Occ')
    doubly_occ = int(dimension_section[enum: enum+50].split()[3])
    enum = dimension_section.find('Doubly Vir')
    doubly_vir = int(dimension_section[enum: enum+50].split()[2])
    enum = dimension_section.find('Frozen Occ')
    frozen_occ = int(dimension_section[enum: enum+50].split()[3])
    enum = dimension_section.find('Frozen Vir')
    frozen_vir = int(dimension_section[enum: enum+50].split()[2])

    enum = dimension_section.find('Total CI configurations')
    total_conf = int(dimension_section[enum: enum+50].split()[3])
    enum = dimension_section.find('Active configurations')
    active_conf = int(dimension_section[enum: enum+50].split()[2])
    enum = dimension_section.find('Hole configurations')
    hole_conf = int(dimension_section[enum: enum+50].split()[2])
    enum = dimension_section.find('Particle configurations')
    particle_conf = int(dimension_section[enum: enum+50].split()[2])

    rasci_dimensions = {'doubly_occupied': doubly_occ,
                        'doubly_virtual': doubly_vir,
                        'frozen_occupied': frozen_occ,
                        'frozen_virtual': frozen_vir,
                        'total_configurations': total_conf,
                        'active_configurations': active_conf,
                        'hole_configurations': hole_conf,
                        'particle_configurations': particle_conf}

    data_dict.update({'rasci_dimensions': rasci_dimensions})

    # Diabatization scheme
    done_diabat = bool(output.find('RASCI DIABATIZATION')+1)
    if done_diabat:
        rot_matrix = _read_simple_matrix('showmatrix final adiabatic -> diabatic', output)[-1]
        adiabatic_matrix = _read_simple_matrix('showing H in adiabatic representation: NO coupling elements', output)[-1]
        diabatic_matrix = _read_simple_matrix('showing H in diabatic representation: WITH coupling elements', output)[-1]

        mulliken_adiabatic = []
        enum = output.find('Mulliken analysis of Adiabatic State')
        for m in re.finditer('Mulliken analysis of Adiabatic State', output[enum:]):
            end_section = search_bars(output, from_position=m.end(), bar_type=r'\-\-\-\-\-\-')[3]
            section_mulliken = output[m.end() + enum: m.end() + enum + end_section]

            #section_mulliken = output[m.end() + enum: m.end() + 20000 + enum]  # 10000: assumed to max of section
            section_mulliken = section_mulliken[:section_mulliken.find('Natural Orbitals stored in FCHK')]
            section_attachment = section_mulliken.split('\n')[9+n_atoms:9+n_atoms*2]

            mulliken_adiabatic.append({'attach': [float(l.split()[1]) for l in section_attachment],
                                       'detach': [float(l.split()[2]) for l in section_attachment],
                                       'total': [float(l.split()[3]) for l in section_attachment]})

        mulliken_diabatic = []
        enum = output.find('showing H in diabatic representation')
        for m in re.finditer('Mulliken Analysis of Diabatic State', output[enum:]):
            end_section = search_bars(output, from_position=m.end(), bar_type=r'\-\-\-\-\-\-')[3]
            section_mulliken = output[m.end() + enum: m.end() + enum + end_section]

            # section_mulliken = output[m.end() + enum: m.end() + 10000 + enum]  # 10000: assumed to max of section
            section_mulliken = section_mulliken[:section_mulliken.find('Natural Orbitals stored in FCHK')]
            section_attachment = section_mulliken.split('\n')[9+n_atoms:9+n_atoms*2]

            mulliken_diabatic.append({'attach': [float(l.split()[1]) for l in section_attachment],
                                      'detach': [float(l.split()[2]) for l in section_attachment],
                                      'total': [float(l.split()[3]) for l in section_attachment]})

        enum = output.find('Transition dipole moment - diabatic states')

        tdm_section = output[enum: enum + 70 * len(rot_matrix)]

        diabatic_tdm = []
        for m in re.finditer('TDM', tdm_section):
            diabatic_tdm.append([float(n) for n in tdm_section[m.end(): m.end()+70].split(':')[1].split()[:3]])

        diabatic_states = []
        for i, tdm in enumerate(diabatic_tdm):
            diabatic_states.append({'excitation_energy': diabatic_matrix[i][i],
                                    'excitation_energy_units': 'eV',
                                    'transition_moment': tdm,
                                    'dipole_moment_units': 'ua',
                                    'mulliken': mulliken_diabatic[i]})

        data_dict['diabatization'] = {'rot_matrix': rot_matrix,
                                      'adiabatic_matrix': adiabatic_matrix,
                                      'diabatic_matrix': diabatic_matrix,
                                      'diabatic_states': diabatic_states,
                                      'mulliken_adiabatic': mulliken_adiabatic}

    # excited states data
    excited_states = []
    for m in re.finditer('RAS-CI total energy for state', output):
        end_section = search_bars(output, from_position=m.start(), bar_type=r'\*'*20)[0]
        section_state = output[m.start():end_section]

        # energies
        enum = section_state.find('total energy for state')
        tot_energy = float(section_state[enum: enum + 50].split()[5])
        enum = section_state.find('Excitation energy')
        exc_energy_units = section_state[enum: enum + 30].split()[2].strip('(').strip(')')
        exc_energy = float(section_state[enum: enum + 80].split()[4])

        # multiplicity
        n_multi = section_state.find('<S^2>')
        multi_data = section_state[n_multi:n_multi + 30].split(':')[1]
        state_multiplicity = float(multi_data.split()[0])

        # dipole moment
        enum = section_state.find('Dipole Moment')
        dipole_mom = [float(section_state[enum:].split()[2]) + 0.0,
                      float(section_state[enum:].split()[4]) + 0.0,
                      float(section_state[enum:].split()[6]) + 0.0]

        # Transition moment
        enum = section_state.find('Trans. Moment')
        trans_mom = strength = None
        if enum > -1:
            trans_mom = [float(section_state[enum:].split()[2]) + 0.0,
                         float(section_state[enum:].split()[4]) + 0.0,
                         float(section_state[enum:].split()[6]) + 0.0]
            trans_mom = standardize_vector(trans_mom)
            strength = float(section_state[enum:].split()[10])

        # Mulliken population analysis
        mulliken_population = None
        enum = section_state.find('Mulliken population analysis')
        if enum > -1:
            max = search_bars(section_state, from_position=enum, bar_type=r'\-'*30)
            pop_charges = []
            pop_spin = []
            for line in section_state[max[1]: max[2]].split('\n')[1:n_atoms+1]:
                c, s = line.split()[2:4]
                pop_charges.append(float(c))
                pop_spin.append(float(s))

            mulliken_population = {'charge': pop_charges, 'spin': pop_spin, 'units': 'au'}

        # Natural orbitals
        nato_occ = None
        enum = section_state.find('NATURAL OCCUPATION NUMBERS')
        if enum > -1:
            lines = []
            for line in section_state[enum:].split('\n')[2::2]:
                if len(line) == 0:
                    break
                if line.split()[0].isnumeric():
                    lines += line.split()[1:]
                else:
                    break
            nato_occ = [float(num) for num in lines]

        # configurations table
        enum = section_state.find('AMPLITUDE')
        enum2 = section_state.find('Contributions')
        section_table = section_state[enum: enum2].split('\n')[2:-2]

        # ' HOLE  | ALPHA | BETA  | PART | AMPLITUDE'
        table = []
        for row in section_table:
            table.append({'hole': row.split('|')[1].strip(),
                          'alpha': row.split('|')[2].strip(),
                          'beta': row.split('|')[3].strip(),
                          'part': row.split('|')[4].strip(),
                          'amplitude': float(row.split('|')[5]) + 0.0})
            table[-1]['occupations'] = get_rasci_occupations_list(table[-1],
                                                                  doubly_occ,
                                                                  basic_data['n_basis_functions'])

        table = sorted(table, key=operator.itemgetter('hole', 'alpha', 'beta', 'part'))
        table = sorted(table, key=lambda x: abs(x['amplitude']), reverse=True)

        # Contributions RASCI wfn
        contributions_section = section_state[enum2:]
        contributions = {'active' : float(contributions_section.split()[4]),
                         'hole': float(contributions_section.split()[6]),
                         'part': float(contributions_section.split()[8])}

        # complete dictionary
        tot_energy_units = 'au'

        state_dict = {'total_energy': tot_energy,
                      'total_energy_units': tot_energy_units,
                      'excitation_energy': exc_energy,
                      'excitation_energy_units': exc_energy_units,
                      'multiplicity': state_multiplicity,
                      'dipole_moment': dipole_mom,
                      'transition_moment': trans_mom,
                      'dipole_moment_units': 'ua',
                      'oscillator_strength': strength,
                      'configurations': table,
                      'contributions_wfn': contributions}

        if nato_occ is not None:
            state_dict.update({'natural_occupation_numbers': nato_occ})

        if mulliken_population is not None:
            state_dict.update({'mulliken_population': mulliken_population})

        excited_states.append(state_dict)

    data_dict.update({'excited_states': excited_states})

    # Interstate transition properties
    done_interstate = bool(output.find('Interstate Transition Properties')+1)
    if done_interstate:
        ini_section = output.find('Interstate Transition Properties')
        end_section = search_bars(output, from_position=ini_section)[1]
        interstate_section = output[ini_section: end_section]

        interstate_dict = {}
        for m in re.finditer('State A: Root', interstate_section):
            section_pair = interstate_section[m.start():m.start() + 10000]
            end_section = search_bars(section_pair, bar_type=r'\*'*20)[0]
            section_pair = section_pair[:end_section]

            lines = section_pair.split('\n')

            state_a = int(lines[0].lower().split('root')[-1])
            state_b = int(lines[1].lower().split('root')[-1])

            pair_dict = {'state_a': state_a,
                         'state_b': state_b}

            s_a = s_b = 0
            for i, line in enumerate(lines):
                # RAS-CI SOC section
                if 'Angular momentum components' in line:
                    pair_dict['angular_momentum'] = [complex(lines[i+1+k].split()[-1].replace('i', 'j')) for k in range(3)]

                if '||gamma^AB||_total' in line:
                    pair_dict['gamma_total'] = float(lines[i+0].split()[-1])
                    pair_dict['gamma_sym'] = float(lines[i+1].split()[-1])
                    pair_dict['gamma_anti_sym'] = float(lines[i+2].split()[-1])

                if "KET: S',Sz'" in line:
                    s_a = float(lines[i].split('=')[1].split()[0])
                    s_b = float(lines[i+1].split('=')[1].split()[0])

                na = int(2 * s_a + 1)
                nb = int(2 * s_b + 1)

                if 'Spin Matrices' in line:
                    warnings.warn('Spin Matrices parsing is deprecated')
                    spinmat_x = _read_soc_matrix(lines[i + 2:], [nb, na])
                    spinmat_y = _read_soc_matrix(lines[i + 4 + nb:], [nb, na])
                    spinmat_z = _read_soc_matrix(lines[i + 6 + 2*nb:], [nb, na])
                    pair_dict['spin_matrices'] = [spinmat_x, spinmat_y, spinmat_z]

                if 'Spin matrices Sx, Sy and Sz for states' in line:
                    warnings.warn('Spin Matrices parsing is deprecated')
                    pair_dict['spin_matrices'] = [np.zeros((nb, na)).tolist()]*3

                if '1-elec SOC matrix (cm-1)' in line:
                    pair_dict['1e_soc_mat'] = _read_soc_matrix(lines[i+1:], [nb, na])

                if '1-elec SOCC' in line:
                    pair_dict['1e_socc'] = float(line.split()[3])

                if '2e-SOMF Reduced matrix elements (cm-1)' in line:
                    r, c = lines[i+1].replace('i', '').split()[-2:]
                    pair_dict['hso_l-'] = float(r) + float(c) * 1j
                    r, c = lines[i+2].replace('i', '').split()[-2:]
                    pair_dict['hso_l0'] = float(r) + float(c) * 1j
                    r, c = lines[i+3].replace('i', '').split()[-2:]
                    pair_dict['hso_l+'] = float(r) + float(c) * 1j

                if '2-elec mean-field SOC matrix (cm-1)' in line:
                    pair_dict['2e_soc_mat'] = _read_soc_matrix(lines[i + 1:], [nb, na])
                if 'Total mean-field SOC matrix (cm-1)' in line:
                    pair_dict['total_soc_mat'] = _read_soc_matrix(lines[i + 1:], [nb, na])
                if 'Mean-Field SOCC' in line:
                    pair_dict['mf_socc'] = float(line.split()[-2])
                    pair_dict['units'] = line.split()[-1]

                if 'Skipping SOCs between states' in line:
                    pair_dict['1e_soc_mat'] = np.zeros((nb, na)).tolist()
                    pair_dict['1e_socc'] = 0.0

                    pair_dict['hso_l-'] = complex(0.0)
                    pair_dict['hso_l0'] = complex(0.0)
                    pair_dict['hso_l+'] = complex(0.0)

                    pair_dict['2e_soc_mat'] = np.zeros((nb, na)).tolist()
                    pair_dict['total_soc_mat'] = np.zeros((nb, na)).tolist()

                    pair_dict['mf_socc'] = 0.0
                    pair_dict['units'] = 'cm-1'

            interstate_dict[(state_a, state_b)] = pair_dict

        _complete_interstate_pairs(interstate_dict)
        data_dict.update({'interstate_properties': interstate_dict})

    # magnetic hyperfine structure
    done_hyperfine = bool(output.find('RAS-CI: MAGNETIC HYPERFINE STRUCTURE')+1)
    if done_hyperfine:
        ini_section = output.find('RAS-CI: MAGNETIC HYPERFINE STRUCTURE')
        end_section = output.find('RAS-CI timing summary (seconds)')
        hyperfine_section = output[ini_section: end_section]

        for i_state, m in enumerate(re.finditer('RAS-CI root: ', hyperfine_section)):
            state_section = hyperfine_section[m.start():m.end()+(n_atoms*1123)]
            # end_section = search_bars(state_section, bar_type='----')[0]

            fermi_contact = []
            spin_dipole = []

            for l, i in enumerate(re.finditer('Nucleus:', state_section)):
                nucleus_section = state_section[i.start():i.end() + 1123]
                end_nucleus_section = search_bars(nucleus_section, bar_type='----')[0]
                nucleus_section = nucleus_section[:end_nucleus_section]
                lines = nucleus_section.split('\n')

                for k, line in enumerate(lines):

                    if 'fermi contact contribution' in line.lower():
                        fermi_contact.append(float(lines[k+1].split()[0]))

                    if 'spin dipole contribution' in line.lower():
                        dipole_x = float(lines[k+1].split()[0])
                        dipole_y = float(lines[k+1].split()[1])
                        dipole_z = float(lines[k+1].split()[2])

                        spin_dipole.append([dipole_x, dipole_y, dipole_z])

            data_dict['excited_states'][i_state].update({'hyperfine_couplings': {'fermi_contact': fermi_contact,
                                                                                 'spin_dipole': spin_dipole}})

    return data_dict



EV_TO_NM = 1239.841  # energy -> wavelength (freq)
AU_TO_EV = 27.211324570273  # wavelength (freq) -> enegy
DEBYE_TO_AU = 0.393430  # dipole moment
ANGSTROM_TO_AU = 1.889716
AMU_TO_ELECTRONMASS = 1822.8884843  # mass
BOHR_TO_CM = 0.00000000529177249
SPEEDOFLIGHT_AU = 137.0359996  # c in au
KB_EV = 8.61735E-05  # Boltzmann constant eV/k
EV_TO_CM = 8065.5401069 # eV to cm-1 (freq)


import numpy as np
from pyqchem.structure import Structure
from pyqchem.errors import ParserError
import re


def basis_format(basis_set_name,
                 atomic_numbers,
                 atomic_symbols,
                 shell_type,
                 n_primitives,
                 atom_map,
                 p_exponents,
                 c_coefficients,
                 p_c_coefficients):
    """
    Function to generate standard basis dictionary

    :param basis_set_name:
    :param atomic_numbers:
    :param atomic_symbols:
    :param shell_type:
    :param n_primitives:
    :param atom_map:
    :param p_exponents:
    :param c_coefficients:
    :param p_c_coefficients:
    :return:
    """

    class TypeList:
        def __getitem__(self, item):
            typeList = {'0': ['s', 1],
                        '1': ['p', 3],
                        '2': ['d', 6],
                        '3': ['f', 10],
                        '4': ['g', 15],
                        '-1': ['sp', 4],
                        '-2': ['d_', 5],
                        '-3': ['f_', 7],
                        '-4': ['g_', 9]}
            if item in typeList:
                return typeList[item]
            else:
                raise ParserError('Basis function of type {} not implemented'.format(item), __name__, '')

    type_list = TypeList()

    atomic_numbers = [int(an) for an in atomic_numbers]
    atom_map = np.array(atom_map, dtype=int)
    # print(atom_map)
    basis_set = {'name': basis_set_name,
                 'primitive_type': 'gaussian'}

    shell_type_index = [0] + np.cumsum([type_list['{}'.format(s)][1]
                                        for s in shell_type]).tolist()
    prim_from_shell_index = [0] + np.cumsum(np.array(n_primitives, dtype=int)).tolist()

    # print(shell_type_index)
    # print(prim_from_shell_index)

    atoms_data = []
    for iatom, atomic_number in enumerate(atomic_numbers):
        symbol = str(atomic_symbols[iatom])

        shell_from_atom_counts = np.unique(atom_map, return_counts=True)[1]
        shell_from_atom_index = np.unique(atom_map, return_index=True)[1]
        # print(shell_from_atom_counts)
        # print('atom_indexes', shell_from_atom_index)
        # print('atom_number', iatom)
        # print('shells index', shell_from_atom_index[iatom])
        # print('number of shells', shell_from_atom_counts[iatom])

        shells_data = []
        for ishell in range(shell_from_atom_counts[iatom]):
            st = type_list['{}'.format(shell_type[shell_from_atom_index[iatom] + ishell])]
            # print(st, ishell)
            ini_prim = prim_from_shell_index[shell_from_atom_index[iatom] + ishell]
            fin_prim = prim_from_shell_index[shell_from_atom_index[iatom] + ishell+1]
            # print(ini_prim)
            # print(fin_prim)

            shells_data.append({
                'shell_type': st[0],
                'functions': st[1],
                'p_exponents': p_exponents[ini_prim: fin_prim],
                'con_coefficients': c_coefficients[ini_prim: fin_prim],
                'p_con_coefficients': p_c_coefficients[ini_prim: fin_prim],
            })

        atoms_data.append({'shells': shells_data,
                           'symbol': symbol,
                           'atomic_number': atomic_number})

    basis_set['atoms'] = atoms_data

    return basis_set


def _reformat_input(array):
    flat_list = []
    for sublist in array:
        for item in sublist:
            if len(item) > 2:
                flat_list.append(item)
            else:
                flat_list.append(item)
    return flat_list


def vect_to_mat(vector):
    n = int(np.sqrt(0.25 + 2 * len(vector)) - 0.5)

    k = 0
    matrix = np.zeros([n, n])
    for i in range(n):
        for j in range(0, i+1):
            matrix[i, j] = vector[k + j]
            matrix[j, i] = vector[k + j]
        k += i+1

    return matrix


def _get_all_nato(output):

    nato_coefficients_list = []
    nato_occupancies_list = []

    for m in re.finditer('Alpha NATO coefficients', output):
        n_elements = int(output[m.end():m.end() + 100].replace('\n', ' ').split()[2])
        nbas = int(np.sqrt(n_elements))
        data = output[m.end(): m.end() + n_elements*50].split()[3:n_elements+3]
        nato_coefficients_list.append({'alpha': np.array(data, dtype=float).reshape(nbas, nbas).tolist()})

    for m in re.finditer('Alpha Natural Orbital occupancies', output):
        n_elements = int(output[m.end():m.end() + 100].replace('\n', ' ').split()[2])
        data = output[m.end(): m.end() + n_elements*50].split()[3:n_elements+3]
        nato_occupancies_list.append({'alpha': np.array(data, dtype=float).tolist()})

    for i, m in enumerate(re.finditer('Beta NATO coefficients', output)):
        n_elements = int(output[m.end():m.end() + 100].replace('\n', ' ').split()[2])
        nbas = int(np.sqrt(n_elements))
        data = output[m.end(): m.end() + n_elements*50].split()[3:n_elements+3]
        nato_coefficients_list[i]['beta'] = np.array(data, dtype=float).reshape(nbas, nbas).tolist()

    for i, m in enumerate(re.finditer('Beta Natural Orbital occupancies', output)):
        n_elements = int(output[m.end():m.end() + 100].replace('\n', ' ').split()[2])
        data = output[m.end(): m.end() + n_elements*50].split()[3:n_elements+3]
        nato_occupancies_list[i]['beta'] = np.array(data, dtype=float).tolist()

    return nato_coefficients_list, nato_occupancies_list


def _get_all_nto(output):

    nto_coefficients_list = []
    nto_occupancies_list = []

    for m in re.finditer('Natural Transition Orbital U coefficients', output):
        n_elements = int(output[m.end():m.end() + 100].replace('\n', ' ').split()[2])
        nbas = int(np.sqrt(n_elements))
        data = output[m.end(): m.end() + n_elements*50].split()[3:n_elements+3]
        nto_coefficients_list.append({'U': np.array(data, dtype=float).reshape(nbas, nbas).tolist()})

    for m in re.finditer('Natural Transition Orbital occupancies', output):
        n_elements = int(output[m.end():m.end() + 100].replace('\n', ' ').split()[2])
        data = output[m.end(): m.end() + n_elements*50].split()[3:n_elements+3]
        nto_occupancies_list.append(np.array(data, dtype=float).tolist())

    for i, m in enumerate(re.finditer('Natural Transition Orbital V coefficients', output)):
        n_elements = int(output[m.end():m.end() + 100].replace('\n', ' ').split()[2])
        nbas = int(np.sqrt(n_elements))
        data = output[m.end(): m.end() + n_elements*50].split()[3:n_elements+3]
        nto_coefficients_list[i]['V'] = np.array(data, dtype=float).reshape(nbas, nbas).tolist()

    return nto_coefficients_list, nto_occupancies_list


def _get_all_fod(output):

    fod_list = []
    for i, m in enumerate(re.finditer('Fractional occupation density', output)):
        n_elements = int(output[m.end():m.end() + 100].replace('\n', ' ').split()[2])
        data = output[m.end(): m.end() + n_elements*50].split()[3:n_elements+3]
        fod_list.append(vect_to_mat(np.array(data, dtype=float)).tolist())

    return fod_list


def _get_all_scf(output):

    scf_list = []
    for i, m in enumerate(re.finditer('Total SCF Density', output)):
        n_elements = int(output[m.end():m.end() + 100].replace('\n', ' ').split()[2])
        data = output[m.end(): m.end() + n_elements*50].split()[3:n_elements+3]
        scf_list.append(vect_to_mat(np.array(data, dtype=float)).tolist())

    return scf_list


def _get_all_nto_new_format(output, label):

    nto_data_dict = {}
    for m in re.finditer('NTOs U coefficients {}'.format(label), output):
        indices = np.array(output[m.end() :m.end() + 100].replace('\n', ' ').split(')')[0].split('(')[1].split(','), dtype=int)
        n_elements = int(output[m.end() :m.end() + 100].replace('\n', ' ').split(')')[1].split()[2])
        nbas = int(np.sqrt(n_elements))
        data = output[m.end(): m.end() + n_elements*50].split(')')[1].split()[3:n_elements+3]
        nto_data_dict[tuple(indices)] = {'U': np.array(data, dtype=float).reshape(nbas, nbas).tolist()}

    for i, m in enumerate(re.finditer('NTOs occupancies {}'.format(label), output)):
        indices = np.array(output[m.end() :m.end() + 100].replace('\n', ' ').split(')')[0].split('(')[1].split(','), dtype=int)
        n_elements = int(output[m.end() :m.end() + 100].replace('\n', ' ').split(')')[1].split()[2])
        data = output[m.end(): m.end() + n_elements*50].split(')')[1].split()[3:n_elements+3]
        nto_data_dict[tuple(indices)]['occupancies'] = np.array(data, dtype=float).tolist()

    for i, m in enumerate(re.finditer('NTOs V coefficients {}'.format(label), output)):
        indices = np.array(output[m.end() :m.end() + 100].replace('\n', ' ').split(')')[0].split('(')[1].split(','), dtype=int)
        n_elements = int(output[m.end() :m.end() + 100].replace('\n', ' ').split(')')[1].split()[2])
        nbas = int(np.sqrt(n_elements))
        data = output[m.end(): m.end() + n_elements*50].split(')')[1].split()[3:n_elements+3]
        nto_data_dict[tuple(indices)]['V'] = np.array(data, dtype=float).reshape(nbas, nbas).tolist()

    return nto_data_dict

def _get_all_tdm(output):

    tdm_data_dict = {}
    for i, m in enumerate(re.finditer('Transition density', output)):
        indices = np.array(output[m.end() :m.end() + 100].replace('\n', ' ').split(')')[0].split('(')[1].split(','), dtype=int)
        n_elements = int(output[m.end() :m.end() + 100].replace('\n', ' ').split(')')[1].split()[2])
        nbas = int(np.sqrt(n_elements))
        data = output[m.end(): m.end() + n_elements*50].split(')')[1].split()[3:n_elements+3]
        tdm_data_dict[tuple(indices)] = np.array(data, dtype=float).reshape(nbas, nbas).tolist()

    return tdm_data_dict


def parser_fchk(output):

    def convert_to_type(item_type, item):
        item_types = {'I': int,
                      'R': float}

        if type(item) is list:
            return [item_types[item_type](e) for e in item]
        else:
            return item_types[item_type](item)

    key_list = ['Charge', 'Multiplicity', 'Number of alpha electrons', 'Number of beta electrons',
                'Atomic numbers', 'Current cartesian coordinates', 'Number of basis functions', 'Shell types',
                'Number of primitives per shell', 'Shell to atom map', 'Primitive exponents',
                'Contraction coefficients', 'P(S=P) Contraction coefficients', 'Alpha MO coefficients',
                'Beta MO coefficients', 'Coordinates of each shell', 'Overlap Matrix',
                'Core Hamiltonian Matrix', 'Alpha Orbital Energies', 'Beta Orbital Energies',
                'Total SCF Density', 'Spin SCF Density', 'Alpha NATO coefficients', 'Beta NATO coefficients',
                'Alpha Natural Orbital occupancies', 'Beta Natural Orbital occupancies',
                'Natural Transition Orbital occupancies', 'Natural Transition Orbital U coefficients',
                'Natural Transition Orbital V coefficients', 'NTOs occupancies SOC', 'NTOs occupancies RAS',
                'Fractional occupation density', 'Transition density',
                ]

    basis_set = output.split('\n')[1].split()[-1]
    words_output = output.replace('\n', ' ').split()

    data = {}
    nw = len(words_output)
    for key in key_list:
        wc = len(key.split())
        for i in range(nw):
            word = ' '.join(words_output[i:i+wc])
            if word == key:
                item_type = words_output[i+wc]
                for l_step in range(5):
                    try:
                        if words_output[l_step + i + wc + 1] == 'N=':
                            item_type = words_output[l_step + i + wc]
                            word += ' '.join(words_output[i + wc:l_step + i + wc])
                            n_elements = int(words_output[l_step + i + wc + 2])
                            data[word] = convert_to_type(item_type, words_output[l_step + i + wc + 3: l_step + i + wc + n_elements + 3])
                        else:
                            item_type = words_output[l_step + i + wc]
                            #print(item_type, words_output[l_step + i + wc + 1])
                            data[word] = convert_to_type(item_type, words_output[l_step + i + wc + 1])
                        break
                    except KeyError:
                        pass
                break

    bohr_to_angstrom = 0.529177249

    coordinates = np.array(data['Current cartesian coordinates']).reshape(-1, 3) * bohr_to_angstrom
    structure = Structure(coordinates=coordinates,
                          atomic_numbers=data['Atomic numbers'],
                          multiplicity=data['Multiplicity'],
                          charge=data['Charge'])

    if not 'P(S=P) Contraction coefficients' in data:
        data['P(S=P) Contraction coefficients'] = np.zeros_like(data['Contraction coefficients']).tolist()

    basis = basis_format(basis_set_name=basis_set,
                         atomic_numbers=structure.get_atomic_numbers(),
                         atomic_symbols=structure.get_symbols(),
                         shell_type=data['Shell types'],
                         n_primitives=data['Number of primitives per shell'],
                         atom_map=data['Shell to atom map'],
                         p_exponents=data['Primitive exponents'],
                         c_coefficients=data['Contraction coefficients'],
                         p_c_coefficients=data['P(S=P) Contraction coefficients'])

    nbas = data['Number of basis functions']

    final_dict = {'structure': structure,
                  'basis': basis,
                  'number_of_electrons': {'alpha': data['Number of alpha electrons'],
                                          'beta': data['Number of beta electrons']}
                  }

    if 'Alpha MO coefficients' in data:
        final_dict['coefficients'] = {'alpha': np.array(data['Alpha MO coefficients']).reshape(-1, nbas).tolist()}
        final_dict['mo_energies'] = {'alpha': data['Alpha Orbital Energies']}

    if 'Beta MO coefficients' in data:
        final_dict['coefficients']['beta'] = np.array(data['Beta MO coefficients']).reshape(-1, nbas).tolist()
        final_dict['mo_energies']['beta'] = data['Beta Orbital Energies']


    if 'Spin SCF Density' in data:
        final_dict['spin_density'] = vect_to_mat(data['Spin SCF Density']).tolist()

    if 'Fractional occupation density' in data:
        final_dict['fractional_occupation_density'] = vect_to_mat(data['Fractional occupation density']).tolist()

    if 'Core Hamiltonian Matrix' in data:
        final_dict['core_hamiltonian'] = vect_to_mat(data['Core Hamiltonian Matrix']).tolist()

    if 'Overlap Matrix' in data:
        final_dict['overlap'] = vect_to_mat(data['Overlap Matrix']).tolist()

    if 'Alpha NATO coefficients' in data:
        final_dict['nato_coefficients'] = {'alpha': np.array(data['Alpha NATO coefficients']).reshape(-1, nbas).tolist()}
        final_dict['nato_occupancies'] = {'alpha': data['Alpha Natural Orbital occupancies']}

    if 'Beta NATO coefficients' in data:
        final_dict['nato_coefficients'].update({
            'beta': np.array(data['Beta NATO coefficients']).reshape(-1, nbas).tolist()})
        final_dict['nato_occupancies'].update({'beta': data['Beta Natural Orbital occupancies']})

    # check multiple NATO (may be improved)
    if 'Alpha NATO coefficients' in data:
        nato_coefficients_list, nato_occupancies_list = _get_all_nato(output)
        if len(nato_occupancies_list) > 1:
            final_dict['nato_coefficients_multi'] = nato_coefficients_list
            final_dict['nato_occupancies_multi'] = nato_occupancies_list

    if 'Natural Transition Orbital occupancies' in data:
        nat_coefficients_list, nat_occupancies_list = _get_all_nto(output)
        if len(nat_occupancies_list) > 1:
            final_dict['nto_coefficients_multi'] = nat_coefficients_list
            final_dict['nto_occupancies_multi'] = nat_occupancies_list

    if 'Fractional occupation density' in data:
        fod_list = _get_all_fod(output)
        if len(fod_list) > 1:
            final_dict['fractional_occupation_density_multi'] = fod_list

    if 'Total SCF Density' in data:
        scf_list = _get_all_scf(output)
        if len(scf_list) > 1:
            final_dict['total_scf_density_multi'] = scf_list
            total = np.array(scf_list[-1])

        else:
            total = np.array(vect_to_mat(data['Total SCF Density']))

        if 'Spin SCF Density' in data:
            spin = np.array(vect_to_mat(data['Spin SCF Density']))
            final_dict['scf_density'] = {'alpha': ((total + spin)/2).tolist(),
                                         'beta': ((total - spin)/2).tolist()}
        else:
            final_dict['scf_density'] = {'alpha':(total/2).tolist(),
                                         'beta': (total/2).tolist()}

        final_dict['total_scf_density'] = total.tolist()

    # Parse new format RAS SOC NTO's
    NTOS_dict = dict(filter(lambda item: "NTOs occupancies SOC" in item[0], data.items()))
    if len(NTOS_dict) > 0:
        final_dict['ntos_soc'] = _get_all_nto_new_format(output, 'SOC')

    # Parse new format RAS NTO's
    NTOS_dict = dict(filter(lambda item: "NTOs occupancies RAS" in item[0], data.items()))
    if len(NTOS_dict) > 0:
        final_dict['ntos_ras'] = _get_all_nto_new_format(output, 'RAS')

    # Parse transisiton density
    TDM_dict = dict(filter(lambda item: "Transition density" in item[0], data.items()))
    if len(TDM_dict) > 0:
        final_dict['transition_density'] = _get_all_tdm(output)

    return final_dict



from pyqchem.structure import Structure
from pyqchem.parsers.common import read_input_structure
import numpy as np
import re


def basic_irc(output, print_data=False):

    branch_mark = True
    data_dict = {}

    # Molecule
    structure = read_input_structure(output)
    n_atoms = structure.get_number_of_atoms()

    forward_steps = []
    backward_steps = []

    list_iterations = [l.end() for l in re.finditer('Reaction path following', output)]
    for ini, fin in zip(list_iterations, list_iterations[1:] + [len(output)]):
        step_section = output[ini:fin]
        enum = step_section.find('Standard Nuclear Orientation')
        atoms_list = step_section[enum:].split('\n')[3:n_atoms+3]
        coordinates_step = np.array([atom.split()[2:] for atom in atoms_list], dtype=float).tolist()

        step_energy = None
        for l in re.finditer('Total energy in the final basis set', step_section):
            step_energy = float(step_section[l.end(): l.end()+50].split()[1])

        step_molecule = Structure(coordinates=coordinates_step,
                                  symbols=structure.get_symbols(),
                                  charge=structure.charge,
                                  multiplicity=structure.multiplicity)

        if (step_section.find('IRC -- maximum number of cycles reached') > 0
                or step_section.find('IRC -- convergence criterion reached') > 0):
            if branch_mark:
                #forward_steps = forward_steps[::-1]
                branch_mark = False
            else:
                break

        if branch_mark:
            forward_steps.append({'molecule': step_molecule,
                                  'energy': step_energy,
                                  })
        else:
            backward_steps.append({'molecule': step_molecule,
                                   'energy': step_energy,
                                   })

    data_dict['irc_forward'] = forward_steps
    data_dict['irc_backward'] = forward_steps

    return data_dict